/*
 * @(#) $Id: plasma2.c 332 2006-01-02 22:42:45Z yeti $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 1996 Stephen Norris
 * Copyright (C) 2001-2002,2004 Yeti (David Necas)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/* XXX  The following version notes are historical and pertain to the
 * XXX  original Plasma plug-in.  See the NEWS file for Plasma2 version
 * XXX  notes.
 */

/*
 * This plug-in produces plasma fractal images. The algorithm is losely
 * based on a description of the fractint algorithm, but completely
 * re-implemented because the fractint code was too ugly to read :)
 *
 * Please send any patches or suggestions to me: srn@flibble.cs.su.oz.au.
 */

/* Version 1.01 */

/*
 * Ported to GIMP Plug-in API 1.0
 *    by Eiichi Takamori <taka@ma1.seikyou.ne.jp>
 *
 * A few functions names and their order are changed :)
 * Plasma implementation almost hasn't been changed.
 *
 * Feel free to correct my WRONG English, or to modify Plug-in Path,
 * and so on. ;-)
 *
 * Version 1.02
 *
 * May 2000
 * tim copperfield [timecop@japan.co.jp]
 * Added dynamic preview mode.
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#else
   /* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#define PROCEDURE_NAME   "plasma2"
#define DATA_KEY_VALS    "plug_in_plasma2"

enum {
    BITS_PER_SAMPLE   = 8,
    CHANNEL_MAX_VALUE = (1U << BITS_PER_SAMPLE) - 1,

    PREVIEW_SIZE = 128,
    SCALE_WIDTH = 128,

    /* FIXME: Is it still impossible to store true unlimited strings as
     * parameters? */
    GRADIENT_NAME_SIZE = 256,
    GRADIENT_SAMPLE_SIZE = 8192,
};

enum {
    RESPONSE_REVERT = 1,
    RESPONSE_RESET
};

typedef enum {
    RANDOM_DISTRIBUTION_UNIFORM = 0,
    RANDOM_DISTRIBUTION_CAUCHY  = 1,
    RANDOM_DISTRIBUTION_EXP     = 2,
    RANDOM_DISTRIBUTION_POW3M   = 3
} RandomDistributionType;

typedef enum {
    SCALING_FUNCTION_EXP   = 0,
    SCALING_FUNCTION_POWER = 1
} FWHMScaleFunction;

typedef enum {
    COLORING_METHOD_INDRGB   = 0,
    COLORING_METHOD_GRAY     = 1,
    COLORING_METHOD_GRADIENT = 2,
    COLORING_METHOD_GREMIX   = 3
} ColoringMethod;

typedef struct {
    gint     seed;
    gint     htil;
    gint     vtil;
    gint     rdt;
    gint     fsf;
    gdouble  lfwhm;
    gdouble  a;
    gint     cm;
    gchar    grad[GRADIENT_NAME_SIZE];
    gboolean random_seed;
} PluginValues;

typedef struct {
    GtkWidget *seed;
    GtkWidget *htil;
    GtkWidget *vtil;
    GtkWidget *rdt;
    GtkWidget *fsf;
    GtkObject *lfwhm;
    GtkObject *a;
    GtkWidget *cm;
    GtkWidget *grad;
    GtkWidget *grad_label;
} PluginControls;

typedef gdouble (*NoisyAverageFunc)(gdouble v1, gdouble v2);

static void    query                      (void);
static void    run                        (const gchar *name,
                                           gint nparams,
                                           const GimpParam *param,
                                           gint *nreturn_vals,
                                           GimpParam **return_vals);
static gint    plugin_dialog              (void);
static void    refresh_controls           (const PluginValues *new_pvals);
static void    random_distribution_changed(GtkWidget *combo,
                                           RandomDistributionType *rdt);
static void    scaling_function_changed   (GtkWidget *combo,
                                           FWHMScaleFunction *fsf);
static void    coloring_method_changed    (GtkWidget *combo,
                                           ColoringMethod *cm);
static void    gradient_changed           (const gchar *name,
                                           gint width,
                                           const gdouble *grad_data,
                                           gboolean dialog_closing,
                                           gpointer user_data);
static void    execute                    (GimpDrawable *drawable);
static void    generate_plasma_channel    (gint xs,
                                           gint ys,
                                           gint channel);
static void    commit                     (GimpDrawable *drawable,
                                           gint channel);
static void    commit_one_tile            (GimpPixelRgn *rgn,
                                           gint channel,
                                           gint gradwidth);
static void    add_transparency           (gint channel);
static void    render_gradient            (void);

static gdouble average_with_noise_uniform (gdouble v1,
                                           gdouble v2);
static gdouble average_with_noise_cauchy  (gdouble v1,
                                           gdouble v2);
static gdouble average_with_noise_exp     (gdouble v1,
                                           gdouble v2);
static gdouble average_with_noise_pow3m   (gdouble v1,
                                           gdouble v2);

static GimpPlugInInfo PLUG_IN_INFO = {
    NULL,  /* init_proc  */
    NULL,  /* quit_proc  */
    query, /* query_proc */
    run,   /* run_proc   */
};

static const PluginValues pvals_defaults = {
    11,                          /* seed        */
    FALSE,                       /* htil        */
    FALSE,                       /* vtil        */
    RANDOM_DISTRIBUTION_UNIFORM, /* rdt         */
    SCALING_FUNCTION_EXP,        /* fsf         */
    0.0,                         /* lfwhm       */
    1.0,                         /* a           */
    COLORING_METHOD_INDRGB,      /* cm          */
    "",                          /* grad        */
    FALSE,                       /* random_seed */
};

static const NoisyAverageFunc noisy_average_funcs[] = {
    &average_with_noise_uniform,
    &average_with_noise_cauchy,
    &average_with_noise_exp,
    &average_with_noise_pow3m,
};

static PluginControls pctrl;
static PluginValues pvals; /* = pvals_defaults; */
static PluginValues pvals_old; /* for revert */

static GRand *rng = NULL;
static GtkWidget *preview_image = NULL;
static GdkPixbuf *preview = NULL;
static gboolean preview_mode;
static gboolean is_rgb;
static GimpDrawable *drawable;
static gdouble *gradient = NULL;
static gint gradient_p_width = -1;
static gboolean in_param_reset = TRUE;

static gint     ix1, iy1, ix2, iy2;     /* Selected image size. */
static gint     bpp, has_alpha, alpha;
static gdouble  max_progress;
static glong    progress;
static gfloat   *plasma; /* plasma channel buffer */
static gdouble  fwhm, randomness; /* true and recomputed fwhm */
static gdouble  xscale, yscale; /* scale (for noise amplitude) */
static NoisyAverageFunc noisy_average = NULL;


MAIN()

static void
query(void)
{
    static GimpParamDef args[] = {
        { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
        { GIMP_PDB_IMAGE, "image", "Input image (unused)" },
        { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
        { GIMP_PDB_INT32, "seed", "Random seed" },
        { GIMP_PDB_INT32, "htil", "Horizontal tileability" },
        { GIMP_PDB_INT32, "vtil", "Vertical tileability" },
        { GIMP_PDB_INT32, "rdt", "Noise distribution type: { UNIFORM (0), CAUCHY (2), EXPONENTIAL (2), CUBIC_TAILS (3) }" },
        { GIMP_PDB_INT32, "fsf", "Noise scaling function: { FRACTINT_LIKE (0), NORRIS_LIKE (1) }" },
        { GIMP_PDB_FLOAT, "lfwhm", "Noise FWHM logarithm" },
        { GIMP_PDB_FLOAT, "a", "Scaling parameter a" },
        { GIMP_PDB_INT32, "cm", "Coloring method: { INDEPENDENT_RGB (0), GRAYSCALE (1), GRADIENT (2), GRADIENT_REMIX (3) }" },
        { GIMP_PDB_STRING, "grad", "Gradient used for coloring (when cm is GRADIENT)" },
    };

    gimp_install_procedure
        (DATA_KEY_VALS,
         "Create a plasma cloud like image to the specified drawable",
         "This plugin creates several types of cloud like images, "
         "mostly variations on the famous `plasma fractal'.",
         "Yeti & authors of original Plasma: "
         "Stephen Norris & (ported to 1.0 by) Eiichi Takamori",
         "Stephen Norris & Yeti",
         "Nov 2002",
         N_("<Image>/Filters/Render/Clouds/Plasma2..."),
         "RGB*, GRAY*",
         GIMP_PLUGIN,
         G_N_ELEMENTS(args), 0, args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam *param,
    gint *nreturn_vals,
    GimpParam  **return_vals)
{
    static GimpParam values[1];
    GimpRunMode run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = param[0].data.d_int32;

    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    textdomain(GETTEXT_PACKAGE);

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    drawable = gimp_drawable_get(param[2].data.d_drawable);
    is_rgb = gimp_drawable_is_rgb(drawable->drawable_id);
    has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);

    pvals = pvals_defaults;
    rng = g_rand_new();
    switch (run_mode) {
        case GIMP_RUN_INTERACTIVE:
        gimp_get_data(DATA_KEY_VALS, &pvals);

        pvals_old = pvals;

        if (!plugin_dialog()) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

        case GIMP_RUN_NONINTERACTIVE:
        if (nparams != 12)
            status = GIMP_PDB_CALLING_ERROR;
        else {
            pvals.seed = (gint)param[3].data.d_int32;
            pvals.htil = (gint)param[4].data.d_int32;
            pvals.vtil = (gint)param[5].data.d_int32;
            pvals.rdt = (RandomDistributionType)param[6].data.d_int32;
            pvals.fsf = (FWHMScaleFunction)param[7].data.d_int32;
            pvals.lfwhm = (gdouble)param[8].data.d_float;
            pvals.a = (gdouble)param[9].data.d_float;
            pvals.cm = (ColoringMethod)param[10].data.d_int32;
            g_strlcpy(pvals.grad, (gchar*)param[11].data.d_string,
                      sizeof(pvals.grad));
        }
        break;

        case GIMP_RUN_WITH_LAST_VALS:
        gimp_get_data(DATA_KEY_VALS, &pvals);
        break;

        default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        if (is_rgb || gimp_drawable_is_gray(drawable->drawable_id)) {
            gimp_progress_init(_("Plasma2..."));

            preview_mode = FALSE;
            in_param_reset = FALSE;
            execute(drawable);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            if (run_mode == GIMP_RUN_INTERACTIVE
                || run_mode == GIMP_RUN_WITH_LAST_VALS) {
                g_strlcpy(pvals.grad, gimp_context_get_gradient(),
                          sizeof(pvals.grad));
                gimp_set_data(DATA_KEY_VALS, &pvals, sizeof(PluginValues));
            }
        }
        else
            status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;
    gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_to_row(GtkWidget *table,
                         GtkWidget *child,
                         gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, 0, GTK_TABLE(table)->ncols,
                     row, row+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline GtkWidget*
yeti_label_new_in_table(const gchar *name,
                        GtkWidget *table,
                        gint row,
                        gboolean sensitive)
{
    GtkWidget *label;

    label = gtk_label_new_with_mnemonic(name);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 0, 1, row, row+1,
                     GTK_EXPAND | GTK_FILL, 0, 0, 0);
    gtk_widget_set_sensitive(label, sensitive);

    return label;
}

static inline GtkWidget*
yeti_check_button_new_with_label(const gchar *name,
                                 gint *value,
                                 GCallback cb,
                                 gpointer data)
{
    GtkWidget *check;

    check = gtk_check_button_new_with_mnemonic(name);
    g_signal_connect(check, "toggled",
                     G_CALLBACK(gimp_toggle_button_update), value);
    g_signal_connect_swapped(check, "toggled", cb, data);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), *value);

    return check;
}

static inline GtkWidget*
yeti_preview_new(GtkWidget *hbox,
                 GimpDrawable *drawable)
{
    GtkWidget *frame, *vbox, *image;
    GdkPixbuf *pixbuf;

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
                            FALSE,
                            BITS_PER_SAMPLE,
                            PREVIEW_SIZE,
                            PREVIEW_SIZE);
    gdk_pixbuf_fill(pixbuf, 0);

    image = gtk_image_new_from_pixbuf(pixbuf);
    g_object_unref(pixbuf);
    gtk_container_add(GTK_CONTAINER(frame), image);

    return image;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            GtkSizeGroup *sgroup,
                            const gint row,
                            gdouble *value,
                            gdouble min,
                            gdouble max,
                            gdouble step,
                            gdouble page,
                            gint digits)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, step, page, digits,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(gimp_double_adjustment_update), value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}
/* }}} */

static gint
plugin_dialog(void)
{
    GtkWidget *dialog, *align, *main_vbox, *hbox, *table;
    GtkSizeGroup *sgroup;
    gint run, row;
    gboolean cm_is_grad, ok;

    preview_mode = TRUE;
    in_param_reset = TRUE;
    gimp_ui_init(PROCEDURE_NAME, TRUE);

    dialog = gimp_dialog_new(_("Plasma2"), PROCEDURE_NAME,
                          NULL, 0,
                          gimp_standard_help_func, "filters/plasma2.html",

                          _("Re_vert"), RESPONSE_REVERT,
                          GIMP_STOCK_RESET, RESPONSE_RESET,
                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                          GTK_STOCK_OK, GTK_RESPONSE_OK,
                          NULL);

    sgroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    main_vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), main_vbox,
                       TRUE, TRUE, 0);

    /* top part with preview and common options */
    hbox = gtk_hbox_new(FALSE, 8);
    gtk_box_pack_start(GTK_BOX(main_vbox), hbox, TRUE, TRUE, 0);

    preview_image = yeti_preview_new(hbox, drawable);
    preview = gtk_image_get_pixbuf(GTK_IMAGE(preview_image));

    table = gtk_table_new(4, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    row = 0;

    align = gtk_alignment_new(0.0, 0.0, 1.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), table);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);

    pctrl.htil = yeti_check_button_new_with_label(_("_Horizontally tileable"),
                                                  &pvals.htil,
                                                  G_CALLBACK(execute),
                                                  drawable);
    yeti_table_attach_to_row(table, pctrl.htil, row);
    row++;

    pctrl.vtil = yeti_check_button_new_with_label(_("_Vertically tileable"),
                                                  &pvals.vtil,
                                                  G_CALLBACK(execute),
                                                  drawable);
    yeti_table_attach_to_row(table, pctrl.vtil, row);
    row++;

    /* gradient, only if rgb */
    if (!is_rgb)
        pvals.cm = COLORING_METHOD_GRAY;
    cm_is_grad = (pvals.cm == COLORING_METHOD_GRADIENT
                  || pvals.cm == COLORING_METHOD_GREMIX);
    pctrl.grad = gimp_gradient_select_widget_new("Plasma2 Gradient", pvals.grad,
                                                 gradient_changed, NULL);
    gtk_table_attach(GTK_TABLE(table), pctrl.grad, 1, 2, row+1, row+2,
                     GTK_FILL, 0, 0, 0);
    gtk_widget_set_sensitive(pctrl.grad, cm_is_grad);

    pctrl.grad_label = yeti_label_new_in_table(_("_Gradient:"),
                                               table, row+1, cm_is_grad);
    gtk_label_set_mnemonic_widget(GTK_LABEL(pctrl.grad_label), pctrl.grad);

    /* coloring function */
    pctrl.cm = gimp_int_combo_box_new(_("Independent RGB"),
                                      COLORING_METHOD_INDRGB,
                                      _("Grayscale"),
                                      COLORING_METHOD_GRAY,
                                      _("Gradient"),
                                      COLORING_METHOD_GRADIENT,
                                      _("Gradient Remix"),
                                      COLORING_METHOD_GREMIX,
                                      NULL);
    gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(pctrl.cm), 1);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, 2,
                              _("Coloring _method:"), 0.0, 0.5,
                              pctrl.cm, 2, TRUE);
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.cm), pvals.cm);
    g_signal_connect(pctrl.cm, "changed",
                     G_CALLBACK(coloring_method_changed), &pvals.cm);
    if (!is_rgb)
        gtk_widget_set_sensitive(pctrl.cm, FALSE);
    row += 2;

    /* generator settings  */
    table = gtk_table_new(5, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_box_pack_start(GTK_BOX(main_vbox), table, FALSE, FALSE, 0);
    row = 0;

    pctrl.seed = gimp_random_seed_new(&pvals.seed, &pvals.random_seed);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, row,
                              _("Random Seed:"), 0.0, 0.5,
                              pctrl.seed, 1, TRUE);
    g_signal_connect_swapped(GIMP_RANDOM_SEED_SPINBUTTON_ADJ(pctrl.seed),
                             "value_changed",
                             G_CALLBACK(execute),
                             drawable);
    row++;

    pctrl.rdt = gimp_int_combo_box_new(_("Uniform"),
                                       RANDOM_DISTRIBUTION_UNIFORM,
                                       _("Cauchy"),
                                       RANDOM_DISTRIBUTION_CAUCHY,
                                       _("Exponential (asymmetric)"),
                                       RANDOM_DISTRIBUTION_EXP,
                                       _("Cubic tails"),
                                       RANDOM_DISTRIBUTION_POW3M,
                                       NULL);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, row,
                              _("Noise _distribution:"), 0.0, 0.5,
                              pctrl.rdt, 1, TRUE);
    gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(pctrl.rdt), 1);
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.rdt), pvals.rdt);
    g_signal_connect(pctrl.rdt, "changed",
                     G_CALLBACK(random_distribution_changed), &pvals.rdt);
    row++;

    pctrl.lfwhm = yeti_scale_entry_new_double(_("Noise _amplitude (log):"),
                                              table, sgroup, row, &pvals.lfwhm,
                                              -7.0, 5.0, 0.01, 0.1, 2);
    g_signal_connect(pctrl.lfwhm, "value-changed",
                     G_CALLBACK(execute), drawable);
    row++;

    pctrl.fsf = gimp_int_combo_box_new(_("Fractint-like,  2^(-a*depth)"),
                                       SCALING_FUNCTION_EXP,
                                       _("Norris-like,  1/depth^a"),
                                       SCALING_FUNCTION_POWER,
                                       NULL);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, row,
                              _("_Scaling function:"), 1.0, 0.5,
                              pctrl.fsf, 2, TRUE);
    gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(pctrl.fsf), 1);
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.fsf), pvals.fsf);
    g_signal_connect(pctrl.fsf, "changed",
                     G_CALLBACK(scaling_function_changed), &pvals.fsf);
    row++;

    pctrl.a = yeti_scale_entry_new_double(_("Scaling _parameter a:"),
                                          table, sgroup, row, &pvals.a,
                                          0.0, 3.0, 0.01, 0.1, 2);
    g_signal_connect(pctrl.a, "value-changed",
                     G_CALLBACK(execute), drawable);
    row++;

    g_object_unref(sgroup);
    gtk_widget_show_all(dialog);
    in_param_reset = FALSE;
    execute(drawable); /* preview image */

    ok = FALSE;
    do {
        run = gimp_dialog_run(GIMP_DIALOG(dialog));
        switch (run) {
            case RESPONSE_REVERT:
            refresh_controls(&pvals_old);
            break;

            case RESPONSE_RESET:
            refresh_controls(&pvals_defaults);
            break;

            default:
            ok = TRUE;
            break;
        }
    } while (!ok);

    gtk_widget_destroy(dialog);

    return run == GTK_RESPONSE_OK;
}

static void
random_distribution_changed(GtkWidget *combo,
                            RandomDistributionType *rdt)
{
    gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(combo), (gint*)rdt);
    execute(drawable);
}

static void
scaling_function_changed(GtkWidget *combo,
                         FWHMScaleFunction *fsf)
{
    gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(combo), (gint*)fsf);
    execute(drawable);
}

static void
coloring_method_changed(GtkWidget *combo,
                        ColoringMethod *cm)
{
    gboolean cm_is_grad;

    gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(combo), (gint*)cm);
    cm_is_grad = (*cm == COLORING_METHOD_GRADIENT
                  || *cm == COLORING_METHOD_GREMIX);
    gtk_widget_set_sensitive(pctrl.grad, cm_is_grad);
    gtk_widget_set_sensitive(pctrl.grad_label, cm_is_grad);
    execute(drawable);
}

static void
gradient_changed(const gchar *name,
                 gint width,
                 const gdouble *grad_data,
                 gboolean dialog_closing,
                 gpointer user_data)
{
    gimp_context_set_gradient(name);
    gradient = g_realloc(gradient, width*sizeof(gdouble));
    memcpy(gradient, grad_data, width*sizeof(gdouble));
    gradient_p_width = width/4;

    execute(drawable);
}

static void
refresh_controls(const PluginValues *new_pvals)
{
    gboolean cm_is_grad;

    in_param_reset = TRUE;

    pvals.htil = new_pvals->htil;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.htil), pvals.htil);

    pvals.vtil = new_pvals->vtil;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.vtil), pvals.vtil);

    pvals.rdt = new_pvals->rdt;
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.rdt), pvals.rdt);

    pvals.fsf = new_pvals->fsf;
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.fsf), pvals.fsf);

    pvals.cm = new_pvals->cm;
    if (!is_rgb)
        pvals.cm = COLORING_METHOD_GRAY;
    cm_is_grad = (pvals.cm == COLORING_METHOD_GRADIENT
                  || pvals.cm == COLORING_METHOD_GREMIX);
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.cm), pvals.cm);
    gtk_widget_set_sensitive(pctrl.grad, cm_is_grad);
    gtk_widget_set_sensitive(pctrl.grad_label, cm_is_grad);

    pvals.lfwhm = new_pvals->lfwhm;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.lfwhm),
                             pvals.lfwhm);

    pvals.a = new_pvals->a;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.a),
                             pvals.a);

    pvals.seed = new_pvals->seed;
    gtk_adjustment_set_value(GIMP_RANDOM_SEED_SPINBUTTON_ADJ(pctrl.seed),
                             pvals.seed);

    in_param_reset = FALSE;
    execute(drawable);
}

/****************************************************************************/

static void
execute(GimpDrawable *drawable)
{
    gint i, n, chmax;

    if (in_param_reset)
        return;

    /* don't use time for seed when rendering preview in full size to get the
       same image as shown in preview */

    /* map lfwhm to true FWHM */
    fwhm = exp(pvals.lfwhm);

    /* compute sizes and other stuff of this kind */
    alpha = -1000; /* to cause a segfault ;-> */
    if (preview_mode) {
        xscale = 1.0/PREVIEW_SIZE;
        yscale = 1.0/PREVIEW_SIZE;
        ix1 = iy1 = 0;
        ix2 = gdk_pixbuf_get_width(preview);
        iy2 = gdk_pixbuf_get_height(preview);
        bpp = gdk_pixbuf_get_n_channels(preview);
    }
    else {
        gimp_drawable_mask_bounds(drawable->drawable_id,
                                  &ix1, &iy1, &ix2, &iy2);
        xscale = 1.0/(ix2 - ix1);
        yscale = 1.0/(iy2 - iy1);
        bpp = gimp_drawable_bpp(drawable->drawable_id);
        if (has_alpha)
            alpha = bpp-1;
    }
    chmax = bpp;

    n = (ix2 - ix1 + 1)*(iy2 - iy1 + 1);
    plasma = g_new(gfloat, n);

    max_progress = n;
    max_progress *= bpp + (pvals.cm == COLORING_METHOD_INDRGB ? bpp : 1);
    max_progress += (has_alpha ? (ix2 - ix1)*(iy2 - iy1) : 0);
    progress = 0;

    noisy_average = noisy_average_funcs[pvals.rdt];
    render_gradient();

    for (i = 0; i < chmax; i++) {
        /* yet more voodoo.  to throw away the duplicate pixels so they don't
           appear on both edges in tileable images, we always generate one pixel
           larger image but pretend it has the right size */
        if (i == 0
            || pvals.cm == COLORING_METHOD_INDRGB
            || pvals.cm == COLORING_METHOD_GREMIX)
            generate_plasma_channel(ix2 - ix1 + 1, iy2 - iy1 + 1, i);

        commit(drawable, i);
    }

    if (has_alpha && preview_mode
        && (pvals.cm == COLORING_METHOD_GRADIENT
            || pvals.cm == COLORING_METHOD_GREMIX)) {
        for (i = 0; i < chmax; i++)
            add_transparency(i);
    }

    if (preview_mode)
        gtk_widget_queue_draw(preview_image);
    else {
        gimp_drawable_flush(drawable);
        gimp_drawable_merge_shadow(drawable->drawable_id, TRUE);
        gimp_drawable_update(drawable->drawable_id,
                             ix1, iy1, (ix2 - ix1), (iy2 - iy1));
    }

    g_free(plasma);
}

/* move one-channel plasma data into preview or drawable */
static void
commit(GimpDrawable *drawable, gint channel)
{
    gpointer pr;
    GimpPixelRgn dest_rgn;

    if (preview_mode) {
        dest_rgn.x = dest_rgn.y = 0;
        dest_rgn.w = dest_rgn.h = PREVIEW_SIZE;
        dest_rgn.bpp = gdk_pixbuf_get_n_channels(preview);
        dest_rgn.rowstride = gdk_pixbuf_get_rowstride(preview);
        dest_rgn.data = gdk_pixbuf_get_pixels(preview);
        commit_one_tile(&dest_rgn, channel, gradient_p_width);
    }
    else {
        gimp_pixel_rgn_init(&dest_rgn, drawable,
                            ix1, iy1, (ix2 - ix1), (iy2 - iy1),
                            TRUE, TRUE);
        /* iterate by tiles */
        for (pr = gimp_pixel_rgns_register(1, &dest_rgn);
             pr != NULL;
             pr = gimp_pixel_rgns_process(pr)) {
            commit_one_tile(&dest_rgn, channel, GRADIENT_SAMPLE_SIZE);
            progress += dest_rgn.w * dest_rgn.h;
            gimp_progress_update(progress/max_progress);
        }
    }
}

/* move one-channel plasma data into one tile of drawable
   when channel == alpha, just fill it with CHANNEL_MAX_VALUE */
static void
commit_one_tile(GimpPixelRgn *rgn, gint channel, gint gradwidth)
{
    gint x, y, w, idx;
    guchar *p;
    gfloat r;

    w = ix2 - ix1 + 1;
    for (y = rgn->y; y < rgn->y + rgn->h; y++) {
        for (x = rgn->x; x < rgn->x + rgn->w; x++) {
            p = rgn->data + (y - rgn->y)*rgn->rowstride + (x - rgn->x)*rgn->bpp;
            if (pvals.cm == COLORING_METHOD_GRADIENT
                || pvals.cm == COLORING_METHOD_GREMIX) {
                idx = plasma[(y - iy1)*w + x - ix1]*(gradwidth - 1);
                r = gradient[4*idx + channel];
            }
            else {
                if (channel == alpha)
                    r = 1.0;
                else
                    r = plasma[(y - iy1)*w + x - ix1];
            }
            p[channel] = CHANNEL_MAX_VALUE*r;
        }
    }
}

static void
add_transparency(gint channel)
{
    static const gint check_colors[] = {
        CHANNEL_MAX_VALUE*GIMP_CHECK_DARK,
        CHANNEL_MAX_VALUE*GIMP_CHECK_LIGHT,
    };

    guchar *p;
    gint x, y, i, w, chc;
    gdouble f;

    w = ix2 - ix1 + 1;
    for (x = ix1; x < ix2; x++) {
        for (y = iy1; y < iy2; y++) {
            i = plasma[(y - iy1)*w + x - ix1]*(gradient_p_width - 1);
            f = gradient[4*i + 3];
            p = gdk_pixbuf_get_pixels(preview)
                + y*gdk_pixbuf_get_rowstride(preview) + x*bpp + channel;
            chc = (x - ix1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE
                + (y - iy1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE;
            *p = (*p)*f + (1.0 - f)*check_colors[chc % 2];
        }
    }
}

static void
render_gradient(void)
{
    gint dummy;

    if (pvals.cm != COLORING_METHOD_GRADIENT
        && pvals.cm != COLORING_METHOD_GREMIX)
        return;

    if (preview_mode) {
        /* don't regenerate what we already have */
        if (gradient == NULL) {
            gradient_p_width = PREVIEW_SIZE;
            gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                              gradient_p_width,
                                              FALSE,
                                              &dummy, &gradient);
        }
    }
    else {
        /* final rendering, done only once at the end */
        g_free(gradient);
        gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                          GRADIENT_SAMPLE_SIZE,
                                          FALSE,
                                          &dummy, &gradient);
    }
}

/* find start and end of block of zeros in flagfield (of size size), starting
   search from i1, the start and end indices are stored in i1 and i2
   returns FALSE when there are no more zeroes in flagfield, TRUE otherwise */
static inline gboolean
find_next_range(gint *i1, gint *i2, const gint *flagfield, gint size)
{
    while (*i1 < size && flagfield[*i1])
        (*i1)++;
    if (*i1 == size)
        return FALSE;
    *i2 = (*i1)--;
    while (*i2 < size && !flagfield[*i2])
        (*i2)++;

    return (*i2) < size;
}

/* math functions */

/* recompute fwhm to distribution-specific parameter */
static inline void
recompute_randomness(double scale, gint w)
{
    static gdouble fwhm_to_rdparam_ratio[] = {
        1, 2, M_LN2, 1,
    };

    g_assert(pvals.rdt < G_N_ELEMENTS(fwhm_to_rdparam_ratio));
    switch (pvals.fsf) {
        case SCALING_FUNCTION_EXP:
        randomness = pow(w*scale, pvals.a);
        break;

        case SCALING_FUNCTION_POWER:
        randomness = pow(-log(w*scale)/M_LN2, -pvals.a);
        break;

        default:
        g_assert_not_reached();
        break;
    }
    randomness *= fwhm/fwhm_to_rdparam_ratio[pvals.rdt];
}

/* the generator */

/* compute and return mean value of v1 and v2 with some random shift
   proportional to randomness (global var) */
static gdouble
average_with_noise_uniform(gdouble v1,
                           gdouble v2)
{
    double val;

    val = 0.5*(v1 + v2);
    val += (-1.0 + 2.0*g_rand_double(rng))*randomness;

    return CLAMP(val, 0.0, 1.0);
}

static gdouble
average_with_noise_cauchy(gdouble v1,
                          gdouble v2)
{
    double val, r;

    val = 0.5*(v1 + v2);
    r = g_rand_double(rng);
    while (G_UNLIKELY(r == 0.5))
        r = g_rand_double(rng);
    val += randomness*tan(G_PI*r);

    return CLAMP(val, 0.0, 1.0);
}

static gdouble
average_with_noise_exp(gdouble v1,
                       gdouble v2)
{
    double val, r;

    val = 0.5*(v1 + v2);
    r = 1.0 - g_rand_double(rng);
    val += -randomness*(log(r) + 1);

    return CLAMP(val, 0.0, 1.0);
}

static gdouble
average_with_noise_pow3m(gdouble v1,
                         gdouble v2)
{
    double val, r, phi;

    val = 0.5*(v1 + v2);
    r = g_rand_double(rng);
    phi = 2.0*G_PI*g_rand_double(rng);
    val += randomness*sin(phi)*sqrt(tan(G_PI*r/2.0));

    return CLAMP(val, 0.0, 1.0);
}

/* one half of a recursion level, horizontal averaging */
static void
horizontal_pass(gint xs,
                gint ys,
                gint depth,
                gint *col_fill,
                gint *col_todo,
                const gint *row_fill,
                gint row_todo)
{
    gint x1, x2, xavg, y, w;

    w = ix2 - ix1 + 1;
    for (x1 = 0; *col_todo > 0; (*col_todo)--) {
        if (!find_next_range(&x1, &x2, col_fill, xs))
            break;
        recompute_randomness(xscale, x2-x1);
        xavg = (x1 + x2)/2;
        col_fill[xavg] = 1;
        for (y = 0; y < ys; y++) {
            if (row_fill[y])
                plasma[y*w + xavg] = noisy_average(plasma[y*w + x1],
                                                   plasma[y*w + x2]);
        }
        if (pvals.vtil)
            plasma[(ys - 1)*w + xavg] = plasma[xavg];
        x1 = x2;
        progress += ys - row_todo;
    }
    if (!preview_mode)
        gimp_progress_update(progress/max_progress);
}

/* one half of a recursion level, vertical averaging */
static void
vertical_pass(gint xs,
              gint ys,
              gint depth,
              gint *row_fill,
              gint *row_todo,
              const gint *col_fill,
              gint col_todo)
{
    gint y1, y2, yavg, x, w;

    w = ix2 - ix1 + 1;
    for (y1 = 0; row_todo > 0; (*row_todo)--) {
        if (!find_next_range(&y1, &y2, row_fill, ys))
            break;
        recompute_randomness(yscale, y2-y1);
        yavg = (y1 + y2)/2;
        row_fill[yavg] = 1;
        for (x = 0; x < xs; x++) {
            if (col_fill[x])
                plasma[yavg*w + x] = noisy_average(plasma[y1*w + x],
                                                   plasma[y2*w + x]);
        }
        if (pvals.htil)
            plasma[yavg*w + xs-1] = plasma[yavg*w];
        y1 = y2;
        progress += xs - col_todo;
    }
    if (!preview_mode)
        gimp_progress_update(progress/max_progress);
}

/* generate single-channel plasma of size (xs,ys), and put it into plasma
   channel is used to generate the right seed value for given channel */
static void
generate_plasma_channel(gint xs,
                        gint ys,
                        gint channel)
{
    gint *row_fill, *col_fill; /* true when slot is filled already  */
    gint col_todo = 0; /* number of empty cols in col_fill */
    gint row_todo = 0; /* number of empty rows in row_fill */
    gint depth; /* `recursion' depth */
    gint x, y, w;

    /* black magic.  we have to use consistent seed value for the preview and
     * for the image for all channels and also consistent behaviour in
     * noninteractive mode.  so we have to reset it for all channels, but
     * obviously to different values. */
    g_rand_set_seed(rng, pvals.seed + 83*channel);

    /* initialize corner pixels
       note we consume the same amount of random numbers regardless of
       tileability */
    w = ix2 - ix1 + 1;
    plasma[0] = g_rand_double(rng);

    plasma[xs - 1] = g_rand_double(rng);
    if (pvals.htil)
        plasma[xs - 1] = plasma[0];

    plasma[(ys - 1)*w] = g_rand_double(rng);
    if (pvals.vtil)
        plasma[(ys - 1)*w] = plasma[0];

    plasma[(ys - 1)*w + xs - 1] = g_rand_double(rng);
    if (pvals.htil)
        plasma[(ys - 1)*w + xs - 1] = plasma[(ys - 1)*w];
    else if (pvals.vtil)
        plasma[(ys - 1)*w + xs - 1] = plasma[xs - 1];
    /* initialize center and edge-center pixels */
    plasma[(ys/2)*w + xs/2] = g_rand_double(rng);
    plasma[xs/2] = g_rand_double(rng);
    plasma[(ys - 1)*w + xs/2] = g_rand_double(rng);
    if (pvals.vtil)
        plasma[(ys - 1)*w + xs/2] = plasma[xs/2];
    plasma[(ys/2)*w] = g_rand_double(rng);
    plasma[(ys/2)*w + xs - 1] = g_rand_double(rng);
    if (pvals.htil)
        plasma[(ys/2)*w + xs - 1] = plasma[(ys/2)*w];

    /* initialize used-slot table */
    col_fill = g_new0(gint, xs);
    row_fill = g_new0(gint, ys);
    col_fill[0] = col_fill[xs-1] = col_fill[xs/2] = 1;
    row_fill[0] = row_fill[ys-1] = row_fill[ys/2] = 1;
    for (x = 0; x < xs; x++)
        col_todo += 1 - col_fill[x];
    for (y = 0; y < ys; y++)
        row_todo += 1 - row_fill[y];

    /* iterate while there remains something to do */
    for (depth = 1; row_todo || col_todo; depth++) {
        /* running vertical and horizontal pass always in the same order would
           create image looking `tall' or `wide', so we alternate them */
        if (depth % 2) {
            vertical_pass(xs, ys, depth,
                          row_fill, &row_todo, col_fill, col_todo);
            horizontal_pass(xs, ys, depth,
                            col_fill, &col_todo, row_fill, row_todo);
        }
        else {
            horizontal_pass(xs, ys, depth,
                            col_fill, &col_todo, row_fill, row_todo);
            vertical_pass(xs, ys, depth,
                          row_fill, &row_todo, col_fill, col_todo);
        }
    }

    g_free(row_fill);
    g_free(col_fill);
}

